/*
 * DBeaver - Universal Database Manager
 * Copyright (C) 2010-2016 Serge Rieder (serge@jkiss.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 2)
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
package org.jkiss.dbeaver.model.struct;

import org.jkiss.code.NotNull;
import org.jkiss.code.Nullable;
import org.jkiss.dbeaver.model.sql.SQLUtils;
import org.jkiss.utils.CommonUtils;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

/**
 * Object filter configuration
 */
public class DBSObjectFilter
{
    private String name;
    private String description;
    private boolean enabled = true;
    private List<String> include;
    private List<String> exclude;

    private transient List<Object> includePatterns = null;
    private transient List<Object> excludePatterns = null;

    public DBSObjectFilter()
    {
    }

    public DBSObjectFilter(String includeString, @Nullable String excludeString)
    {
        if (include != null) {
            this.include = SQLUtils.splitFilter(includeString);
        }
        if (exclude != null) {
            this.exclude = SQLUtils.splitFilter(excludeString);
        }
    }

    public DBSObjectFilter(DBSObjectFilter filter)
    {
        if (filter != null) {
            this.name = filter.name;
            this.description = filter.description;
            this.enabled = filter.enabled;
            this.include = filter.include == null ? null : new ArrayList<>(filter.include);
            this.exclude = filter.exclude == null ? null : new ArrayList<>(filter.exclude);
        }
    }

    public String getName()
    {
        return name;
    }

    public void setName(String name)
    {
        this.name = name;
    }

    public String getDescription()
    {
        return description;
    }

    public void setDescription(String description)
    {
        this.description = description;
    }

    public boolean isEnabled()
    {
        return enabled;
    }

    public void setEnabled(boolean enabled)
    {
        this.enabled = enabled;
    }

    public List<String> getInclude()
    {
        return include;
    }

    public void addInclude(String name)
    {
        if (include == null) {
            include = new ArrayList<>();
        }
        include.add(name);
        this.includePatterns = null;
    }

    public void setInclude(List<String> include)
    {
        this.include = include;
        this.includePatterns = null;
    }

    public List<String> getExclude()
    {
        return exclude;
    }

    public void addExclude(String name)
    {
        if (exclude == null) {
            exclude = new ArrayList<>();
        }
        exclude.add(name);
        this.excludePatterns = null;
    }

    public void setExclude(List<String> exclude)
    {
        this.exclude = exclude;
        this.excludePatterns = null;
    }

    public boolean isNotApplicable()
    {
        return !enabled || isEmpty();
    }

    public boolean isEmpty()
    {
        return CommonUtils.isEmpty(include) && CommonUtils.isEmpty(exclude);
    }

    public boolean hasSingleMask()
    {
        return include != null && include.size() == 1 && CommonUtils.isEmpty(exclude);
    }

    @Nullable
    public String getSingleMask()
    {
        return !CommonUtils.isEmpty(include) ? include.get(0) : null;
    }
    
    public synchronized boolean matches(String name)
    {
        if (includePatterns == null && !CommonUtils.isEmpty(include)) {
            includePatterns = new ArrayList<>(include.size());
            for (String inc : include) {
                if (!inc.isEmpty()) {
                    includePatterns.add(makePattern(inc));
                }
            }
        }
        if (includePatterns != null) {
            // Match includes (at least one should match)
            boolean matched = false;
            for (Object pattern : includePatterns) {
                if (matchesPattern(pattern, name)) {
                    matched = true;
                    break;
                }
            }
            if (!matched) {
                return false;
            }
        }

        if (excludePatterns == null && !CommonUtils.isEmpty(exclude)) {
            excludePatterns = new ArrayList<>(exclude.size());
            for (String exc : exclude) {
                if (!exc.isEmpty()) {
                    excludePatterns.add(makePattern(exc));
                }
            }
        }
        if (excludePatterns != null) {
            // Match excludes
            for (Object pattern : excludePatterns) {
                if (matchesPattern(pattern, name)) {
                    return false;
                }
            }
        }
        // Done
        return true;
    }

    private static boolean matchesPattern(Object pattern, String name) {
        if (pattern instanceof Pattern) {
            return ((Pattern)pattern).matcher(name).matches();
        } else {
            return ((String)pattern).equalsIgnoreCase(name);
        }
    }

    @NotNull
    private static Object makePattern(String str) {
        if (SQLUtils.isLikePattern(str)) {
            return Pattern.compile(
                SQLUtils.makeLikePattern(str), Pattern.CASE_INSENSITIVE | Pattern.MULTILINE);
        } else {
            return str;
        }
    }

}
